from sklearn.gaussian_process import GaussianProcessRegressor
import numpy as np
from typing import Callable
import numpy.typing as npt
from typing import Optional
import inspect
from gpalexp.gpr_fit_predict import gpr_fit, gpr_predict, next_design



def gpal_optimize(gpr:GaussianProcessRegressor, 
                  num_features: int, 
                  data_record:npt.NDArray[np.floating],
                  stimulus_candidates:npt.NDArray[np.floating], 
                  stimulus_masking_function:Optional[Callable] = None, 
                  return_stdev: bool = True, 
                  return_covar:bool = False):
    if not isinstance(gpr, GaussianProcessRegressor):
        raise TypeError(f"gpr should be a GaussianProcessRegressor instance, got {type(gpr).__name__}.")
    if not isinstance(num_features, int):
        raise TypeError(f"num_features should be an integer value, got the type of {type(num_features).__name__}.")
    if num_features<1:
        raise ValueError(f"num_features should be a positive integer, got {num_features}.")
    if not isinstance(data_record, np.ndarray):
        raise TypeError(f"data_record should be a numpy array, got the type of {type(data_record).__name__}.")
    if data_record.dtype!=np.floating:
        raise TypeError(f"data_record should have the float dtype, got the dtype of {data_record.dtype}.")
    if data_record.ndim!=2:
        raise ValueError(f"data_record should be a 2D array, got {data_record.ndim} dimensions.")
    if data_record.shape[1]!=num_features+1:
        raise ValueError(f"data_record should have {num_features+1} columns; got {data_record.shape[1]} columns.")
    if not isinstance(stimulus_candidates, np.ndarray):
        raise TypeError(f"stimulus_candidates should be a numpy array, got the type of {type(stimulus_candidates).__name__}.")
    if stimulus_candidates.dtype!=np.floating:
        raise TypeError(f"stimulus_candidates should have the float dtype, got the dtype of {stimulus_candidates.dtype}.")
    if stimulus_candidates.ndim!=2:
        raise ValueError(f"stimulus_candidates should be a 2D array, got {stimulus_candidates.ndim} dimensions.")
    if stimulus_candidates.shape[1]!=num_features:
        raise ValueError(f"stimulus_candidates should have {num_features} columns, got {stimulus_candidates.shape[1]}.")
    if stimulus_masking_function is not None:
        if not callable(stimulus_masking_function):
            raise TypeError(f"stimulus_masking_function should be a callable function, got the type of {type(stimulus_masking_function).__name__}.")
        masking_function_params_num=len(inspect.signature(stimulus_masking_function).parameters)
        if masking_function_params_num != num_features:
            raise ValueError(f"stimulus_masking_function should have {num_features} parameters, got {masking_function_params_num} parameters.")
    if not isinstance(return_stdev, bool):
        raise TypeError(f"return_stdev should be a bool value, got the type of {type(return_stdev).__name__}.")
    if not isinstance(return_covar, bool):
        raise TypeError(f"return_covar should be a bool value, got the type of {type(return_covar).__name__}.")    


    if stimulus_masking_function is None:
        predict_candidates_X=stimulus_candidates
    else:
        stimulus_candidates_T=stimulus_candidates.T
        stimulus_mask_binary=stimulus_masking_function(*stimulus_candidates_T) 
        predict_candidates_X=stimulus_candidates_T[:,stimulus_mask_binary].T

    lml=gpr_fit(gpr=gpr, num_features=num_features, data_record=data_record)
    posterior_prediction=gpr_predict(gpr, num_features=num_features, predict_candidates_X=predict_candidates_X,
                                    return_stdev=return_stdev, return_covar=return_covar)
    result, gp_mean, gp_std=next_design(posterior_prediction=posterior_prediction, num_features=num_features,
                                                            predict_candidates_X=predict_candidates_X)

    if num_features==1:
        result=result[0]
    return result, gp_mean, gp_std, lml



from psychopy import visual, event
from psychopy.visual import ElementArrayStim
from sklearn.gaussian_process.kernels import Kernel
import numpy as np
import math, time, random


def initialize_psychopy(fullscr=False):
    """
    Initialize PsychoPy visual elements and return dictionary of them.
    
    Args: None
    Returns: Dictionary of PsychoPy visual components
    """
    win = visual.Window([1800, 1200], color='grey', units='pix', fullscr=fullscr)
    line = visual.Line(win, start=(-500, 0), end=(500, 0), lineColor='black')
    line_leftend = visual.Line(win, start=(-500, -10), end=(-500, 10), lineColor='black')
    line_rightend = visual.Line(win, start=(500, -10), end=(500, 10), lineColor='black')
    left_box = visual.Rect(win, width=300, height=300, pos=(-500, -200), fillColor=None, lineColor='black')
    right_box = visual.Rect(win, width=300, height=300, pos=(500, -200), fillColor=None, lineColor='black')
    question_box = visual.Rect(win, width=300, height=300, pos=(0, 200), fillColor=None, lineColor='black')
    marker = visual.Line(win, start=(0, -10), end=(0, 10), lineColor='orange', lineWidth=3)
    img_stim = visual.ImageStim(
        win=win,
        image='./images/random_noise.png',
        pos=(0, 200),
        size=(300, 300)
    )
    visuals = {'win': win,
                'line': line,
                'line_leftend': line_leftend,
                'line_rightend': line_rightend,
                'left_box': left_box,
                'right_box': right_box,
                'question_box': question_box,
                'marker': marker,
                'img_stim': img_stim
                }   

    return visuals

def draw_grid_position(n_dots, dot_size, boxW, boxH, box_center=(0, 0), padding=10):
    """
    Draws a grid of positions for the dots within a specified box, 
    so that the dots are evenly spaced.

    Args:
    - n_dots: number of dots
    - dot_size: size(diameter) of dots
    - boxW, boxH: dimensions of the box (width, height)
    - box_center: center position of the box (x, y)
    - padding: padding between dots and box edges

    Returns:
    - jittered_positions: list of tuples representing the (x, y) positions of the dots
    """
    itemWH = dot_size * 1.75
    usable_boxW = boxW - 2 * (padding + dot_size / 2)
    usable_boxH = boxH - 2 * (padding + dot_size / 2)

    X = int(usable_boxW // itemWH)
    Y = int(usable_boxH // itemWH)

    if X * Y < n_dots:
        raise ValueError("Not enough grid positions for the number of dots.")
    
    grid_x = np.linspace(-usable_boxW / 2, usable_boxW / 2, X)
    grid_y = np.linspace(-usable_boxH / 2, usable_boxH / 2, Y)
    grid_X, grid_Y = np.meshgrid(grid_x, grid_y)
    grid_positions = list(zip(grid_X.flatten(), grid_Y.flatten()))
    random.shuffle(grid_positions)

    selected_positions = grid_positions[:n_dots]

    jittered_positions = []
    jitter_ratio = 0.3
    for x, y in selected_positions:
        jitter_x = random.uniform(-itemWH * jitter_ratio, itemWH * jitter_ratio)
        jitter_y = random.uniform(-itemWH * jitter_ratio, itemWH * jitter_ratio)
        final_x = x + jitter_x + box_center[0]
        final_y = y + jitter_y + box_center[1]
        jittered_positions.append((final_x, final_y))

    return jittered_positions

def calculate_dot_size(max_number_size, number, max_number=500):
    """
    Calculate the size of the dots based on the number of dots.
    Size of the given number of dots is calculated to be same as the cumulative area of the upperbound dots.

    Args:
    - max_number_size: maximum size(diameter) of the dots
    - number: number of dots
    - max_number: maximum number of dots

    Returns:
    - dot_size: size(diameter) of the dots
    """
    original_r = max_number_size / 2
    total_upper_area = math.pi * (original_r ** 2) * max_number

    current_dot_area = total_upper_area / number
    current_r = math.sqrt(current_dot_area / math.pi)

    stimdot_size1 = math.floor(current_r * 2)
    stimdot_size2 = math.ceil(current_r * 2)

    error1 = abs(total_upper_area - (stimdot_size1 / 2) ** 2 * math.pi * number)
    error2 = abs(total_upper_area - (stimdot_size2 / 2) ** 2 * math.pi * number)

    stimdot_size = stimdot_size1 if error1 < error2 else stimdot_size2
    return stimdot_size

def draw_base_components(visuals):
    """
    Draw base components (line, left_box, right_box, leftend, rightend, question_box) on win.

    Args:
    - win: PsychoPy window object
    - visuals: Dictionary of PsychoPy visual components
    """

    visuals['left_box'].draw()
    visuals['right_box'].draw()
    visuals['question_box'].draw()
    visuals['line'].draw()
    visuals['line_leftend'].draw()
    visuals['line_rightend'].draw()

def show_and_get_response(number, visuals, max_number, size_control=False):
    """
    Show and get response from the user.

    Args:
    - number: number of dots
    - visuals: dictionary of PsychoPy visual components to show
    - max_number: maximum number of dots
    - size_control: whether to control the size of the dots

    Returns:
    - response: the user's response
    """
    win = visuals['win']
    marker = visuals['marker']
    img_stim = visuals['img_stim']

    if isinstance(number, float):
        number=int(number)
    if isinstance(max_number, float):
        max_number=int(max_number)
    if size_control:
        dot_size = calculate_dot_size(max_number_size=4, number=number, max_number=max_number)
    else:
        dot_size = 4

    # Draw dots & base visual components
    right_dots_pos = draw_grid_position(max_number, 4, 300, 300, box_center=(500, -200), padding=25)
    right_dots = ElementArrayStim(win,
                                   nElements=max_number,
                                   xys=right_dots_pos,
                                   sizes=4,
                                   colors='orange',
                                   elementTex=None,
                                   elementMask='circle',
                                   interpolate=True)

    question_dots_pos = draw_grid_position(number, dot_size, 300, 300, box_center=(0, 200), padding=25)
    question_dots = ElementArrayStim(win,
                                      nElements=number,
                                      xys=question_dots_pos,
                                      sizes=dot_size,
                                      colors='orange',
                                      elementTex=None,
                                      elementMask='circle',
                                      interpolate=True)
    draw_base_components(visuals)
    right_dots.draw()
    question_dots.draw()
    win.flip()
    start_time = time.time()

    # Defining mouse click flag
    clicked = False
    
    while not clicked:
        if time.time() - start_time > 2:
            # After 2 seconds, mask the given dots box
            img_stim.draw()
            right_dots.draw()
            draw_base_components(visuals)
            win.flip()

        # Waiting for mouse click
        mouse = event.Mouse(win=win)
        if mouse.getPressed()[0]:
            x, y = mouse.getPos()

            # If the click is outside the valid area, ignore it
            if x < -500 or x > 500 or y < -20 or y > 20:
                continue

            # If the click is valid, record the position and show the marker
            clicked = True
            marker.pos = (x, 0)
            marker.draw()
            win.flip()

    # Draw the response position
    visuals['line'].draw()
    visuals['line_leftend'].draw()
    visuals['line_rightend'].draw()
    visuals['left_box'].draw()
    visuals['right_box'].draw()
    right_dots.draw()
    marker.draw()
    win.flip()

    # Return the participant's response (estimation value)
    return int((x + 500) / 1000 * max_number)



from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Kernel, ConstantKernel, DotProduct
from sklearn.gaussian_process.kernels import ExpSineSquared, Exponentiation, Matern
from sklearn.gaussian_process.kernels import PairwiseKernel, RBF, RationalQuadratic, WhiteKernel
from sklearn.gaussian_process.kernels import Product, Sum
import numpy as np
from functools import reduce
from typing import Optional
from typing import Union
import numpy.typing as npt
import re


def GPRInstance(kernel_types:list[str], 
                kernel_arguments:list[dict], 
                combine_format:str, 
                alpha:Union[float, npt.NDArray[np.floating]]=1e-10, 
                normalize_y:bool=True, 
                n_restarts_optimizer:int=50, 
                random_state:Optional[Union[int, np.random.RandomState]]=None):
    if not isinstance(combine_format, str):
        raise TypeError(f"combine_format should be a string, got {type(format).__name__}.")
    if not isinstance(alpha, Union[float, np.ndarray]):
        raise TypeError(f"alpha should be float or numpy array, got {type(alpha).__name__}.")
    if isinstance(alpha, float):
        if alpha<=0:
            raise ValueError(f"alpha should be a positive value, got {alpha}.")
    if isinstance(alpha, np.ndarray):
        if alpha.dtype!=np.floating:
            raise TypeError(f"alpha should have the float dtype.")
        if not np.all(alpha>0):
            raise ValueError(f"alpha should only contain positive elements.")
    if not isinstance(normalize_y, bool):
        raise TypeError(f"normalize_y should be a bool value, got {type(normalize_y).__name__}.")
    if not isinstance(n_restarts_optimizer, int):
        raise TypeError(f"n_restarts_optimizer should be an integer value, got {type(n_restarts_optimizer).__name__}.")
    if n_restarts_optimizer<0:
        raise ValueError(f"n_restarts_optimizer should be non-negative, got {n_restarts_optimizer}.")
    if random_state is not None:
        if not (isinstance(random_state, int) or isinstance(random_state, np.random.RandomState)):
            raise TypeError(f"random_state should be None, an int value, or a RandomState object. Got {type(random_state).__name__}")


    kernel_builder=KernelBuilder(num_kernels=len(kernel_types), 
                           kernel_types_strs_list=kernel_types, 
                           kernel_arguments_dics_list=kernel_arguments, 
                           combine_format_str=combine_format)
    created_kernel=kernel_builder.create_compound_kernel()
    gpr=GaussianProcessRegressor(kernel=created_kernel, alpha=alpha, normalize_y=normalize_y, 
                                 n_restarts_optimizer=n_restarts_optimizer, random_state=random_state)
    return created_kernel, gpr


## Adding and multiplying kernel instances are processed after creating all individual kernel instances.
class KernelBuilder():
    def __init__(self, 
                 num_kernels:int, 
                 kernel_types_strs_list:list[str], 
                 kernel_arguments_dics_list:list[dict], 
                 combine_format_str:str):
        if not isinstance(num_kernels, int):
            raise TypeError(f"num_kernels should be an int value, got the type of {type(num_kernels).__name__}.")
        if num_kernels<=0:
            raise ValueError(f"num_kernels should be a positive integer, got {num_kernels}.")
        if not isinstance(kernel_types_strs_list, list):
            raise TypeError(f"kernel_types_str_list should be a list, got the type of {type(kernel_types_strs_list).__name__}.")
        if not all(isinstance(kernel, str) for kernel in kernel_types_strs_list):
            typ=kernel_types_strs_list[[isinstance(kernel, str) for kernel in kernel_types_strs_list].index(False)]
            raise TypeError(f"kernel_types_str_list should contain string elements, got a {type(typ).__name__ } type element.")
        if len(kernel_types_strs_list)!=num_kernels:
            raise ValueError(f"kernel_types_str_list should be of length {num_kernels}, got {len(kernel_types_strs_list)}.")
        if not isinstance(kernel_arguments_dics_list, list):
            raise TypeError(f"kernel_arguments_dics_list should be a list, got the type of {type(kernel_arguments_dics_list).__name__}.")
        if not all(isinstance(kad, dict) for kad in kernel_arguments_dics_list):
            typ=kernel_arguments_dics_list[[isinstance(kad, dict) for kad in kernel_arguments_dics_list].index(False)]
            raise TypeError(f"kernel_arguments_dics_list should contain dictionary elements, got a {type(typ).__name__} type element.")
        if len(kernel_arguments_dics_list)!=num_kernels:
            raise ValueError(f"kernel_arguments_dics_list should be of length {num_kernels}, got {len(kernel_arguments_dics_list)}.")
        if not isinstance(combine_format_str, str):
            raise TypeError(f"combine_format_str should be a string, got the type of {type(combine_format_str).__name__}.")
        
        split_format_elements=re.split(r'(\+|\*)', combine_format_str)
        if len(split_format_elements) != 2*num_kernels-1:
            raise ValueError(f"combine_format_str should indicate a valid combination of kernel objects.")
        for i in range(0, (len(split_format_elements)+1)//2):
            if int(split_format_elements[2*i][1:])!=i+1:
                raise ValueError(f"The {i}-th operand of combine_format_str should be k{i+1}, not {split_format_elements[2*i]}.")
            
        split_format_numbers=re.findall(r'\d+', combine_format_str)
        kernel_symbol_idxs=list(map(int, split_format_numbers))
        if max(kernel_symbol_idxs)!=num_kernels:
            raise ValueError(f"combine_format_str should include {num_kernels} kernel symbols (from k1 to k{num_kernels}), got a symbol k{max(kernel_symbol_idxs)}.")
        
        split_format_unalloweds=re.findall(r'[^a-zA-z0-9+*]', combine_format_str)
        if split_format_unalloweds:
            raise ValueError(f"combine_format_str should only include alphabet characters, numbers, +, and *, got {split_format_unalloweds}.")
        

        self.num_kernels=num_kernels
        self.kernel_types_strs_list=kernel_types_strs_list
        self.kernel_arguments_dics_list=kernel_arguments_dics_list
        self.split_format_elements=split_format_elements

        self.basicTypes=['ConstantKernel', 'DotProduct', 'ExpSineSquared', 'Exponentiation',
                        'Matern', 'PairwiseKernel', 'RBF', 'RationalQuadratic', 'WhiteKernel']
        self.final_kernel: Optional[Kernel] = None
        self.individual_kernel_objects=[]
    
    def create_compound_kernel(self):
        if self.num_kernels==1:
            final_kernel=self.create_individual_kernel(kernel_type_str=self.kernel_types_strs_list[0], 
                                                       kernel_args_dict=self.kernel_arguments_dics_list[0])    
            self.final_kernel=final_kernel
        else:
            for kernel_index in range(self.num_kernels):
                kernel_object=self.create_individual_kernel(kernel_type_str=self.kernel_types_strs_list[kernel_index], 
                                                            kernel_args_dict=self.kernel_arguments_dics_list[kernel_index])
                self.individual_kernel_objects.append(kernel_object)

            for elem_index, format_elem in enumerate(self.split_format_elements):
                if format_elem=="*":  # * is always at the odd index
                    operand_first=self.individual_kernel_objects[(elem_index-1)//2]
                    operand_second=self.individual_kernel_objects[(elem_index+1)//2]
                    kernel_product=operand_first*operand_second
                    self.individual_kernel_objects.insert((elem_index-1)//2, kernel_product)
                    self.individual_kernel_objects.remove(operand_first)
                    self.individual_kernel_objects.remove(operand_second)
                else:
                    continue
            final_compound_kernel=reduce(lambda x,y:x+y, self.individual_kernel_objects)
            self.final_kernel=final_compound_kernel
        return self.final_kernel
    
    def create_individual_kernel(self, 
                                 kernel_type_str:str, 
                                 kernel_args_dict:dict):
                                 
        def check_params_match(kernel:Kernel, given_args_dict:dict):
            kernel_params=kernel.get_params().keys()
            if not set(given_args_dict.keys()).issubset(set(kernel_params)):
                raise ValueError(f"The set of parameters provided ({set(given_args_dict.keys())}) \nis not the part of the set of required parameters: \n{set(kernel_params)}.")
            return True

        if not isinstance(kernel_type_str, str):
            raise TypeError(f"kernel_type_str should be a string value, got the type of {type(kernel_type_str).__name__}.")
        if not isinstance(kernel_args_dict, dict):
            raise TypeError(f"kernel_args_dict should be a dictionary, got the type of {type(kernel_args_dict).__name__}.")
        
        kernel_type_index=0
        kernel_object=None
        try:
            kernel_type_index=self.basicTypes.index(kernel_type_str)
        except ValueError:
            raise ValueError(f"The given kernel type '{kernel_type_str}' is not a valid kernel type.")
        
        if kernel_type_index==0:
            kernel_object=ConstantKernel()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)

        elif kernel_type_index==1:
            kernel_object=DotProduct()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)
        
        elif kernel_type_index==2:
            kernel_object=ExpSineSquared()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)

        elif kernel_type_index==3:
            if len(kernel_args_dict.keys())!=2:
                raise ValueError(f"The Exponentiation kernel takes 2 arguments, got {len(kernel_args_dict.keys())} arguments.")
            if not isinstance(kernel_args_dict['kernel'], Kernel):
                raise TypeError(f'''The 'kernel' argument of the Exponentiation kernel should be a valid kernel instance, 
                                got the type of {type(kernel_args_dict['kernel']).__name__}.''')
            if not isinstance(kernel_args_dict['exponent'], float):
                raise TypeError(f'''The 'exponent' argument of the Exponentiation kernel should be a float value, 
                                got the type of {type(kernel_args_dict['exponent']).__name__}.''')
            kernel_object=Exponentiation(**kernel_args_dict)

        elif kernel_type_index==4:
            kernel_object=Matern()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)
        
        elif kernel_type_index==5:
            kernel_object=PairwiseKernel()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)

        elif kernel_type_index==6:
            kernel_object=RBF()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)

        elif kernel_type_index==7:
            kernel_object=RationalQuadratic()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)

        elif kernel_type_index==8:
            kernel_object=WhiteKernel()
            if check_params_match(kernel=kernel_object, given_args_dict=kernel_args_dict):
                kernel_object.set_params(**kernel_args_dict)
        
        else:
            raise ValueError(f"The type '{kernel_type_str}' is not a valid kernel type.")

        return kernel_object



from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import *
import numpy as np
from functools import reduce
from typing import Optional
import numpy.typing as npt
from gpalexp.utils import prediction

def gpr_fit(gpr:GaussianProcessRegressor, 
            num_features: int, 
            data_record: npt.NDArray[np.floating]):
    if not isinstance(gpr, GaussianProcessRegressor):
        raise TypeError(f"gpr should be a GaussianProcessRegressor object, got the type of {type(gpr).__name__}.")
    if not isinstance(num_features, int):
        raise TypeError(f"num_features should be an integer value, got the type of {type(num_features).__name__}.")
    if num_features<1:
        raise ValueError(f"num_features should be a positive integer, got {num_features}.")
    if not isinstance(data_record, np.ndarray):
        raise TypeError(f"data_record should be a numpy array.")
    if data_record.dtype!=np.floating:
        raise TypeError(f"data_record should have the float dtype, got the dtype of {data_record.dtype}.")
    if data_record.ndim!=2:
        raise ValueError(f"data_record should be a 2D numpy array, got {data_record.ndim} dimensions.")
    if data_record.shape[1]!=num_features+1:
         raise ValueError(f"data_record should be a 2D numpy array with {num_features+1} columns, got {data_record.shape[1]} columns.")
    
    fit_data_X=data_record[:, :-1]
    obs_data_Y=data_record[:, -1]
    gpr.fit(X=fit_data_X, y=obs_data_Y)
    log_marginal_likelihood=gpr.log_marginal_likelihood()
    return log_marginal_likelihood


def gpr_predict(gpr:GaussianProcessRegressor, 
                num_features:int, 
                predict_candidates_X: npt.NDArray[np.floating], 
                return_stdev:bool=False, 
                return_covar:bool=False):
    if not isinstance(gpr, GaussianProcessRegressor):
        raise TypeError(f"gpr should be a GaussianProcessRegressor object, got the type of {type(gpr).__name__}.")
    if not isinstance(num_features, int):
        raise TypeError(f"num_features should be an integer value, got the type of {type(num_features).__name__}.")
    if num_features<1:
        raise ValueError(f"num_features should be a positive integer, got {num_features}.")
    if return_stdev and return_covar:
        raise ValueError(f"At most one of return_stdev and return_covar can be True.")
    if not isinstance(predict_candidates_X, np.ndarray):
        raise TypeError(f"predict_data_X should be a numpy array, got the type of {type(predict_candidates_X).__name__}.")
    if predict_candidates_X.dtype!=np.floating:
        raise TypeError(f"predict_data_X should have the float dtype.")
    if predict_candidates_X.ndim!=2:
        raise ValueError(f"predict_data_X should be a 2D numpy array, got {predict_candidates_X.ndim} dimensions.")
    if predict_candidates_X.shape[1]!=num_features:
         raise ValueError(f"predict_data_X should be a 2D numpy array with {num_features} columns, got {predict_candidates_X.shape[1]} columns.")
    
    posterior_mean=None
    posterior_stdev: Optional[np.ndarray] = None
    posterior_covariance: Optional[np.ndarray] = None
    if return_stdev and not return_covar:
        posterior_mean, posterior_stdev=gpr.predict(X=predict_candidates_X, 
                                                    return_std=return_stdev, 
                                                    return_cov=return_covar)
    if not return_stdev and return_covar:
        posterior_mean, posterior_covariance=gpr.predict(X=predict_candidates_X, 
                                                         return_std=return_stdev, 
                                                         return_cov=return_covar)
    if not return_stdev and not return_covar:
        posterior_mean = gpr.predict(X=predict_candidates_X, 
                                     return_std=return_stdev, 
                                     return_cov=return_covar)
    
    posterior_prediction_per_candidate=prediction(mean=posterior_mean, std=posterior_stdev, cov=posterior_covariance)
    return posterior_prediction_per_candidate


def next_design(posterior_prediction:prediction, 
                num_features: int, 
                predict_candidates_X:npt.NDArray[np.floating]):
    
    if not isinstance(posterior_prediction, prediction):
        raise TypeError(f"posterior_prediction should be the value returned from gpr_predict() function.")
    if not isinstance(num_features, int):
        raise TypeError(f"num_features should be an integer value, got the type of {type(num_features).__name__}.")
    if num_features<1:
        raise ValueError(f"num_features should be a positive integer, got {num_features}.")
    if not isinstance(predict_candidates_X, np.ndarray):
        raise TypeError(f"predict_candidates_X should be a numpy array, got the type of {type(predict_candidates_X).__name__}.")
    if predict_candidates_X.dtype!=np.floating:
        raise TypeError(f"predict_candidates_X should have the float dtype.")
    if predict_candidates_X.ndim!=2:
        raise ValueError(f"predict_candidates_X should be a 2D array, got {predict_candidates_X.ndim} dimensions.")
    if predict_candidates_X.shape[1]!=num_features:
        raise ValueError(f"predict_candidates_X should be a 2D numpy array with {num_features} columns, got {predict_candidates_X.shape[1]} columns.")    
    if posterior_prediction.mean is None:
        raise ValueError(f"The 'mean' field of posterior_predictions should not be None.")
    posterior_mean=posterior_prediction.mean
    if posterior_prediction.std is not None:
        posterior_stdev=posterior_prediction.std
    if posterior_prediction.cov is not None:
        posterior_covar=posterior_prediction.cov

    candidate_idx_with_max_stdev=np.argmax(posterior_stdev)
    next_stimulus_coordinate=predict_candidates_X[candidate_idx_with_max_stdev]
    next_feature_posterior_mean=posterior_mean[candidate_idx_with_max_stdev].item()
    next_feature_posterior_stdev=posterior_stdev[candidate_idx_with_max_stdev].item()
    return next_stimulus_coordinate.tolist(), next_feature_posterior_mean, next_feature_posterior_stdev



import argparse

def argparser():
    parser = argparse.ArgumentParser(description='Run the Number Line Estimation Experiment.')
    parser.add_argument('--num_trials', type=int, default=20, help='Number of trials in the experiment.')
    parser.add_argument('--seed', default=None, help='Random seed for reproducibility.')
    parser.add_argument('--num_features', type=int, default=1, help='The number of stimulus features to be optimized.')
    
    ## Arguments related to gpr_instance.py
    parser.add_argument('--kernel_types', type=list, default=[0,6,8], help='The indices (or names) of the kernels to be combined.')
    parser.add_argument('--kernel_arguments', type=list, default=[[1.0], [1.0], [0.01]], help='The values to be fed to create each kernel objects.')
    parser.add_argument('--combine_format', type=str, default="k1*k2+k3", help='A string representing how the kernels should be combined.')
    parser.add_argument('--normalize_y', default=True, help='A binary mask indicating whether to normalize obs_data_Y while fitting.')
    parser.add_argument('--n_restarts_optimizer', default=0, help='The number of restarts of the optimizer to find the optimal kernel parameters.')
    parser.add_argument('--gpr_random_state', default=None, help='A parameter determining random number generation in initializing the centers.')
    
    ## Arguments related to gpr_optimize.py
    parser.add_argument('--return_std', default=True, help='A binary mask indicating whether to return standard deviation of posterior distribution at each query value.')
    parser.add_argument('--return_cov', default=False, help='A binary mask indicating whether to return covaraince matrix of posterior distribution at each query value.')

    parser.add_argument('--save_results_dir', default='results', help='A directory to store the task results in .csv format.')
    parser.add_argument('--save_models_dir', default='models', help='A directory to store the trained Gaussian process regressor model.')
    parser.add_argument('--save_figures_dir', default='figures', help='A directory to store the figures for analyzing the results.')
    
    
    ## return args
    return parser.parse_args()


from typing import Optional, Tuple

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.ticker import MultipleLocator
from mpl_toolkits.mplot3d.axes3d import Axes3D
from sklearn.metrics import mean_squared_error
from sklearn.gaussian_process import GaussianProcessRegressor

def plot_GPAL_uncertainty(fig_size:Tuple[int, int], 
                          fit_data_X:npt.NDArray, 
                          obs_data_Y:npt.NDArray, 
                          predict_candidates_X:npt.NDArray, 
                          post_mean:npt.NDArray, 
                          post_stdev:npt.NDArray, 
                          x_label:str, 
                          y_label:str, 
                          title:str, 
                          sigma_coef:float=1.0):
    """
    This function plots the uncertainty plot.
    This 'uncertainty plot' visualizes the experimental data as a scatterplot.
    Then it visualizes the posterior mean calculated for each stimulus candidate,
    and the uncertainty range determined by the associated posterior standard deviations.

    - Parameter Descriptions
    fig_size: The size of the figure. Must be a tuple holding integer values.
    fit_data_X: The numpy array recording the provided stimulus features for each given experimental trial. 
    obs_data_Y: The numpy array recording the subject's responses for each given experimental trial.
    predict_candidates_X: The numpy array containing the stimulus candidates for GPAL optimization.
    post_mean: The numpy array holding the posterior mean values for each stimulus candidate (in predict_candidate_X).
    post_stdev: The numpy array holding the posterior standard deviation value for each stimulus candidate (in predict_candidate_X).
    x_label: The text label associated with the x-axis (stimulus candidates).
    y_label: The text label associated with the y-axis (subject responses).
    title: The title of the figure.
    sigma_coef: The coefficient determining the uncertainty range. Must be a positive float value.

    NOTE: The uncertainty range is defined as the following:
          [post_mean - sigma_coef * post_stdev, post_mean + sigma_coef * post_stdev]
    NOTE: The post_mean and post_stdev must be the posterior statistics 
          obtained by gpr.predict(predict_candidates_X) after gpr.fit(fit_data_X, obs_data_Y).
          In other words, post_mean and post_stdev must hold posterior statistics
          for GPAL optimization of the upcoming (next) trial.
          Then GPAL will select the stimulus candidate with the largest posterior standard deviation,
          which is equivalent to the largest uncertainty range.

    Return Value Descriptions
    figure: The resulting figure.
    ax: A plot visualized in the resulting figure.
    """
    if not isinstance(fig_size, Tuple):
        raise TypeError(f"fig_size should be a tuple, got the type of {type(fig_size).__name__}.")
    if any([not isinstance(fs, int) for fs in fig_size]):
        raise ValueError(f"fig_size should contain integer elements.")
    if len(fig_size)!=2:
        raise ValueError(f"figsize should be of length 2, got {len(fig_size)}.")
    if not isinstance(fit_data_X, np.ndarray):
        raise TypeError(f"fit_data_X should be a numpy array, got the type of {type(fit_data_X).__name__}.")
    if fit_data_X.ndim!=2:
        raise ValueError(f"fit_data_X should be a 2D array, got {fit_data_X.ndim} dimensions.")
    if not isinstance(predict_candidates_X, np.ndarray):
        raise TypeError(f"predict_candidates_X should be a numpy array, got the type of {type(predict_candidates_X).__name__}.")
    if predict_candidates_X.ndim!=1:
        raise ValueError(f"predict_candidates_X should be a 1D array, got {predict_candidates_X.ndim} dimensions.")
    if not isinstance(obs_data_Y, np.ndarray):
        raise TypeError(f"obs_data_Y should be a numpy array, got the type of {type(obs_data_Y).__name__}.")
    if obs_data_Y.ndim!=1:
        raise ValueError(f"obs_data_Y should be a 1D array, got {obs_data_Y.ndim} dimensions.")
    if not isinstance(post_mean, np.ndarray):
        raise TypeError(f"post_mean should be a numpy array, got the type of {type(post_mean).__name__}.")
    if post_mean.ndim!=1:
        raise ValueError(f"post_mean should be a 1D array, got {post_mean.ndim} dimensions.")
    if not isinstance(post_stdev, np.ndarray):
        raise TypeError(f"post_stdev should be a numpy array, got {type(post_stdev).__name__}.")
    if post_stdev.ndim!=1:
        raise ValueError(f"post_stdev should be a 1D array, got {post_stdev.ndim} dimensions.")
    if fit_data_X.shape[0]!=obs_data_Y.shape[0]:
        raise ValueError(f"fit_data_X and obs_data_Y should have equal number of data, got {fit_data_X.shape[0]} and {obs_data_Y.shape[0]}.")
    if predict_candidates_X.shape[0]!=post_mean.shape[0]:
        raise ValueError(f"predict_candidates_X and post_mean should have equal number of data, got {predict_candidates_X.shape[0]} and {post_mean.shape[0]}.")
    if predict_candidates_X.shape[0]!=post_stdev.shape[0]:
        raise ValueError(f"predict_candidates_X and post_stdev should have equal number of data, got {predict_candidates_X.shape[0]} and {post_stdev.shape[0]}.")
    if not isinstance(sigma_coef, float):
        raise TypeError(f"sigma_coef should be a float value, got the type of {type(sigma_coef).__name__}.")
    if sigma_coef<0:
        raise ValueError(f"sigma_coef should be non-negative, got {sigma_coef}.")
    if not isinstance(x_label, str):
        raise TypeError(f"xlabel should be a string value, got the type of {type(x_label).__name__}.")
    if not isinstance(y_label, str):
        raise TypeError(f"ylabel should be a string value, got the type of {type(y_label).__name__}.")
    if not isinstance(title, str):
        raise TypeError(f"title should be a string value, got the type of {type(title).__name__}.")
    
    figure=plt.figure(figsize=fig_size)
    ax=figure.add_subplot(1,1,1)
    
    ## Plots the experiment data as a scatterplot.
    ax.scatter(fit_data_X.ravel(), obs_data_Y, c='black', label='Data')
    ## Plots the posterior mean values associated with every stimulus candidate.
    ax.plot(predict_candidates_X, post_mean, label="Prediction", linewidth=2.5, color='black')
    ## Plots the uncertainty range with semi-transparent color.
    ax.fill_between(predict_candidates_X, post_mean-sigma_coef*post_stdev, 
                     post_mean+sigma_coef*post_stdev, alpha=0.3, label='Uncertainty')
    
    ## Setting the title and labels.
    ax.set_title(title, fontsize=28)
    ax.set_xlabel(x_label, fontsize=24)
    ax.set_ylabel(y_label, fontsize=24)
    
    return figure, ax



def plot_GPAL_compare_uncertainty(fig_size:Tuple[int, int], 
                                  font_size:int, 
                                  fit_data_X:npt.NDArray, 
                                  obs_data_Y:npt.NDArray,  
                                  predict_candidates_X:npt.NDArray, 
                                  post_mean_previous:npt.NDArray, 
                                  post_stdev_previous:npt.NDArray, 
                                  post_mean_target:npt.NDArray, 
                                  post_stdev_target:npt.NDArray,
                                  xlabel:str,
                                  ylabel:str,
                                  title:str, 
                                  title_previous:str, 
                                  title_target:str, 
                                  max_stdev_design:float, 
                                  sigma_coef:float=1.0):
    """
    This function plots the uncertainty plot at two consecutive trials.
    Since GPAL optimizes the experiment design adaptively, 
    we can visualize how the uncertainty plot changes after a single iteration of GPAL.
    This figure visualizes the uncertainty plot at a certain 'target' trial (on the right subplot) 
    and at the 'previous' trial (on the left subplot).
    The uncertainty is calculated based on the GPAL optimization conducted up to the (target/previous) trial.
    We can directly observe how the uncertainty interval shrinks at the GPAL-selected design point.
    Moreover, a dotted vertical line indicates the stimulus candidate value
    whose associated posterior standard deviation is the largest.
    We can see the design coordinate of the new experiment design (a red dot in the target trial)
    exactly corresponds to the position of the vertical line in the previous trial.

    - Parameter descriptions
    fig_size: The size of the figure. Must be a tuple holding integer values.
    font_size: The font size of the text in the figure. Must be positive.
    fit_data_X: The numpy array recording the provided design variables, up to the target trial
    obs_data_Y: The numpy array recording the subject responses, up to the target trial
    predict_candidates_X: The numpy array containing the stimulus candidates for GPAL optimization.
    post_mean_previous: The numpy array holding the posterior mean value for each stimulus candidate, up to the previous trial
    post_stdev_previous: The numpy array holding the posterior standard deviation value for each stimulus candidate, up to the previous trial.
    post_mean_target: The numpy array holding the posterior mean values, up to the target trial.
    post_stdev_target: The numpy array holding the posterior standard deviation values, up to the target trial.
    title: The title of the whole figure.
    title_previous: The title of the left figure (GPAL up to the previous trial)
    title_target: The title of the right figure (GPAL up to the target trial)
    max_stdev_design: The stimulus feature value with maximum posterior standard deviation, in the previous trial.
    sigma_coef: The coefficient determining the uncertainty range. Must be a positive float value.

    - Return Value Specifications
    figure: The whole resulting figure.
    ax1: A plot visualized in the left subplot of the figure.
    ax2: A plot visualized in the right subplot of the figure.
    """
    if not isinstance(fig_size, tuple):
        raise TypeError(f"fig_size should be a tuple, got the type of {type(fig_size).__name__}")
    if any([not isinstance(fs, int) for fs in fig_size]):
        raise ValueError(f"fig_size should contain integer elements.")
    if len(fig_size)!=2:
        raise ValueError(f"fig_size should be of length 2, got {len(fig_size)}.")
    if not isinstance(font_size, int):
        raise TypeError(f"font_size should be an integer value, got the type of {type(font_size).__name__}.")
    if font_size<=0:
        raise ValueError(f"font_size should be a positive value, got {font_size}.")
    if not isinstance(fit_data_X, np.ndarray):
        raise TypeError(f"fit_data_X should be a numpy array, got the type of {type(fit_data_X).__name__}.")
    if fit_data_X.ndim!=2:
        raise ValueError(f"fit_data_X should be a 2D array, got {fit_data_X.ndim} dimensions.")
    if not isinstance(predict_candidates_X, np.ndarray):
        raise TypeError(f"predict_candidates_X should be a numpy array, got the type of {type(predict_candidates_X).__name__}.")
    if predict_candidates_X.ndim!=1:
        raise ValueError(f"predict_candidates_X should be a 1D array, got {predict_candidates_X.ndim} dimensions.")
    if not isinstance(obs_data_Y, np.ndarray):
        raise TypeError(f"obs_data_Y should be a numpy array, got the type of {type(obs_data_Y).__name__}.")
    if obs_data_Y.ndim!=1:
        raise ValueError(f"obs_data_Y should be a 1D array, got {obs_data_Y.ndim} dimensions.")
    if not isinstance(post_mean_previous, np.ndarray):
        raise TypeError(f"post_mean_previous should be a numpy array, got the type of {type(post_mean_previous).__name__}.")
    if post_mean_previous.ndim!=1:
        raise ValueError(f"post_mean_previous should be a 1D array, got {post_mean_previous.ndim} dimensions.")
    if not isinstance(post_stdev_previous, np.ndarray):
        raise TypeError(f"post_stdev_previous should be a numpy array, got the type of {type(post_stdev_previous).__name__}.")
    if post_stdev_previous.ndim!=1:
        raise ValueError(f"post_stdev_previous should be a 1D array, got {post_stdev_previous.ndim} dimensions.")
    if not isinstance(post_mean_target, np.ndarray):
        raise TypeError(f"post_mean_target should be a numpy array, got the type of {type(post_mean_target).__name__}.")
    if post_mean_target.ndim!=1:
        raise ValueError(f"post_mean_target should be a 1D array, got {post_mean_target.ndim} dimensions.")
    if not isinstance(post_stdev_target, np.ndarray):
        raise TypeError(f"post_stdev_target should be a numpy array, got the type of {type(post_stdev_target).__name__}.")
    if post_stdev_target.ndim!=1:
        raise ValueError(f"post_stdev_target should be a 1D array, got {post_stdev_target.ndim} dimensions.")
    if fit_data_X.shape[0]!=obs_data_Y.shape[0]:
        raise ValueError(f"fit_data_X and obs_data_Y should have equal number of data, got {fit_data_X.shape[0]} and {obs_data_Y.shape[0]}.")
    if predict_candidates_X.shape[0]!=post_mean_previous.shape[0]:
        raise ValueError(f"predict_candidates_X and post_mean_previous should have equal number of data, got {predict_candidates_X.shape[0]} and {post_mean_previous.shape[0]}.")
    if predict_candidates_X.shape[0]!=post_stdev_previous.shape[0]:
        raise ValueError(f"predict_candidates_X and post_stdev_previous should have equal number of data, got {predict_candidates_X.shape[0]} and {post_stdev_previous.shape[0]}.")
    if predict_candidates_X.shape[0]!=post_mean_target.shape[0]:
        raise ValueError(f"predict_candidates_X and post_mean_target should have equal number of data, got {predict_candidates_X.shape[0]} and {post_mean_target.shape[0]}.")
    if predict_candidates_X.shape[0]!=post_stdev_target.shape[0]:
        raise ValueError(f"predict_candidates_X and post_stdev_target should have equal number of data, got {predict_candidates_X.shape[0]} and {post_stdev_target.shape[0]}.")
    if not isinstance(max_stdev_design, float):
        raise TypeError(f"max_stdev_design should be a float value, got the type of {type(max_stdev_design).__name__}.")
    if not isinstance(sigma_coef, float):
        raise TypeError(f"sigma_coef should be a float value, got the type of {type(sigma_coef).__name__}")
    if sigma_coef<0:
        raise ValueError(f"sigma_coef should be non-negative, got {sigma_coef}.")
    if not isinstance(xlabel, str):
        raise TypeError(f"xlabel should be a string value, got the type of {type(xlabel).__name__}.")
    if not isinstance(ylabel, str):
        raise TypeError(f"ylabel should be a string value, got the type of {type(ylabel).__name__}.")
    if not isinstance(title, str):
        raise TypeError(f"title should be a string value, got the type of {type(title).__name__}.")
    if not isinstance(title_previous, str):
        raise TypeError(f"title_previous should be a string value, got the type of {type(title_previoius).__name__}.")
    if not isinstance(title_target, str):
        raise TypeError(f"title_target should be a string value, got the type of {type(title_target).__name__}.")
    
    ## Creating a figure with two subplots.
    figure, (ax1, ax2)=plt.subplots(2,1, figsize=fig_size)
    
    ## Left subplot
    ## Plotting the experiment data up to the previous trial.
    ax1.scatter(fit_data_X[:-1], obs_data_Y[:-1], c='black', label='Data')
    ## Plotting the posterior mean, calculated with so-far obtained experiment data.
    ax1.plot(predict_candidates_X, post_mean_previous, label="Prediction", linewidth=2.5, color='black')
    ## Plotting the uncertainty range for each design candidate.
    ax1.fill_between(predict_candidates_X.ravel(), post_mean_previous-sigma_coef*post_stdev_previous,
                     post_mean_previous+sigma_coef*post_stdev_previous, alpha=0.3, label='Uncertainty')
    ## Plotting a dotted vertical line, at the design candidate associated with maximum posterior standard deviation.
    ax1.axvline(x=fit_data_X[-1], color='green', linestyle='--', linewidth=3)
    ax1.set_xlabel(xlabel, fontsize=16)
    ax1.set_ylabel(ylabel, fontsize=16)
    ## Setting the title for the left subplot.
    ax1.set_title(title_previous, fontsize=font_size)
    

    ## Right subplot
    ## Plotting the experiment data up to the target trial.
    ax2.scatter(fit_data_X, obs_data_Y, c='black')
    ## Plotting the posterior mean, calculated with so-far obtained experiment data.
    ax2.plot(predict_candidates_X, post_mean_target, linewidth=2.5, color='black')
    ## Plotting the uncertainty range for each design candidate.
    ax2.fill_between(predict_candidates_X.ravel(), post_mean_target-sigma_coef*post_stdev_target, 
                     post_mean_target+sigma_coef*post_stdev_target, alpha=0.3)
    ## Plotting an experiment data for the newly selected experimental design.
    ax2.scatter(fit_data_X[-1], obs_data_Y[-1], c='red')
    ## Plotting a dotted vertical line, at the design candidate associated with maximums posterior standard deviation.
    ax2.axvline(x=max_stdev_design, color='green', linestyle='--', linewidth=3)
    ax2.set_xlabel(xlabel, fontsize=16)
    ax2.set_ylabel(ylabel, fontsize=16)
    ## Setting the title for the right subplot.
    ax2.set_title(title_target, fontsize=font_size)

    ## Setting the title for the whole figure.
    figure.suptitle(title, fontsize=font_size)
    plt.tight_layout()
    
    return figure, ax1, ax2




def plot_frequency_histogram_1D(fig_size:Tuple[int, int], 
                                num_data:int, 
                                stimulus_feature:npt.NDArray, 
                                bins:int, 
                                ranges:Optional[Tuple[float, float]], 
                                x_label:str, 
                                y_label:str, 
                                title:str, 
                                mode:str="sum"):
    """
    plot_frequency_histogram_1D plots the stimulus selection frequencies of 1D GPAL as a 2D histogram.
    GPAL selects the optimal stimulus among stimulus candidates in an adaptive manner.
    We can observe the distribution of those 'selected' stimulus values,
    and therefore examine properties of the function of our interest.
    As "1D" in the function name implies, this functions plots a 2D histogram,
    where the x-axis denotes the stimulus values and the y-axis indicate the frequencies.

    - Parameter Descriptions.
    fig_size: The size of the figure. Must be a tuple holding integer values.
    num_data: The number of selected stimuli (i.e. the optimal stimulus candidates).  
    design_var: A numpy array holding all selected stimuli.
    bins: The number of equal-length bins dividing the range of selected stimuli.
    ranges: A tuple indicating the total range of the selected stimuli.
    x_label: The text label associated with the x-axis (stimulus candidates).
    y_label: The text label associated with the y-axis (subject responses).
    title: The title of the figure.
    mode: A string determining the mode of the histogram. Must be either 'sum' or 'average'.
          Setting it to 'average' will return the relative frequencies, which will be summed to 1.

    NOTE: The ranges parameter will be automatically set to the folloiwng, if not specified explicitly.
          ranges = (np.min(stimulus_feature)), np.max(stimulus_feature))

    - Return Value Specifications
    figure: The whole resulting figure.
    ax: A plot visualized in the figure.
    """
    if not isinstance(fig_size, tuple):
        raise TypeError(f"fig_size should be a tuple, got the type of {type(fig_size).__name__}.")
    if any([not isinstance(fs, int) for fs in fig_size]):
        raise TypeError(f"fig_size should have integer elements.")
    if len(fig_size)!=2:
        raise ValueError(f"fig_size should be of length 2, got {len(fig_size)}.")
    if not isinstance(num_data, int):
        raise TypeError(f"num_data should be an integer value, got the type of {type(num_data).__name__}.")
    if num_data<1:
        raise ValueError(f"num_data should be a positive integer, got {num_data}.")
    if not isinstance(stimulus_feature, np.ndarray):
        raise TypeError(f"stimulus_feature should be a numpy array, got the type of {type(stimulus_feature).__name__}.")
    if stimulus_feature.ndim!=1:
        raise ValueError(f"stimulus_feature should be a 1D array, got {stimulus_feature.ndim} dimensions.")
    if not isinstance(bins, int):
        raise TypeError(f"bins should be an integer value, got the type of {type(bins).__name__}.")
    if ranges is not None:
        if not isinstance(ranges, tuple):
            raise TypeError(f"ranges should be a tuple or None, got the type of {type(ranges).__name__}.")
        if not all([isinstance(r, float) for r in ranges]):
            raise TypeError(f"ranges should contain float elements.")
        if len(ranges)!=2:
            raise ValueError(f"ranges should be of length 2, got {len(ranges)}.")
    if not isinstance(mode, str):
        raise TypeError(f"mode should be a string value, got the type of {type(mode).__name__}.")
    if mode not in ["average", "sum"]:
        raise ValueError(f"mode should be either 'average' or 'sum', got {mode}.")
    if not isinstance(x_label, str):
        raise TypeError(f"x_label should be a string value, got the type of {type(x_label).__name__}.")
    if not isinstance(y_label, str):
        raise TypeError(f"y_label should be a string value, got the type of {type(y_label).__name__}.")
    if not isinstance(title, str):
        raise TypeError(f"title should be a string value, got the type of {type(title).__name__}.")

    ## Drawing a figure
    figure=plt.figure(figsize=fig_size)
    ax=figure.add_subplot(1,1,1)

    ## Creating a histogram with np.nistogram()
    ## hist: The values of the resulting histogram.
    ## dv1_pos: The values at the edge of each bins. 
    hist, dv1_pos=np.histogram(stimulus_feature, bins=bins, range=ranges)
    if mode=='average':
        hist=hist/num_data

    ## Determining the width of each bar in the figure.
    bin_width = (ranges[1] - ranges[0]) / bins if ranges else dv1_pos[1] - dv1_pos[0]
    ## Determining the position (x-coordinate) of the center of each bar.
    bin_centers = dv1_pos[:-1] + bin_width / 2
    
    ## Drawing a bar plot with the obtained histogram values
    ax.bar(
        x=bin_centers,
        height=hist.ravel(),
        width=bin_width,
        bottom=0,
        align='center',
        color='skyblue',       
        alpha=0.6,             
        edgecolor='black',     
        linewidth=0.8          
    )

    ## Setting the labels and the title.
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title(title)

    return figure, ax


def plot_frequency_histogram_2D(fig_size:Tuple[int, int], 
                                num_data:int, 
                                stimulus_feature:npt.NDArray, 
                                bins:list[int], 
                                ranges:Optional[list[list[float]]], 
                                xlabel:str, 
                                ylabel:str, 
                                zlabel:str, 
                                title:str, 
                                mode:str='sum'):
    """
    This function plots the stimulus selection frequencies of 2D GPAL as a 3D histogram.
    GPAL selects the optimal stimulus among stimulus candidates in an adaptive manner.
    We can observe the distribution of those 'selected' stimuli,
    therefore examine features of the function of our interest.
    As "2D" in the function name implies, this functions plots a 3D histogram,
    where the (x,y) coordinate represents the stimulus and the z-axis indicates the frequencies.


    - Parameter Descriptions.
    fig_size: The size of the figure. Must be a tuple holding integer values.
    num_data: The number of selected stimuli (i.e. the optimal ones).  
    design_var: A numpy array holding all selected stimuli. 
                Each row corresponds to a single stimulus. Must contain 2 columns (due to 2D GPAL).
    bins: A list holding the number of equal-length bins for the x,y axis. 
          Each element corresponds to the number of bins for each axis. Must be of length 2.
    ranges: A list of 2 lists, indicating the total range of the selected stimuli.
            The first list elements specifies the range of the first stimulus feature of the selected stimuli.
            The second list element specifies the range of the second stimulus feature of the selected stimuli.
    x_label: The text label associated with the x-axis (first stimulus feature)
    y_label: The text label associated with the y-axis (second stimulus feature)
    z_label: The text label associated with the z-axis.
    title: The title of the figure.
    mode: A string determining the mode of the histogram. Must be either 'sum' or 'average'.
          Setting it to 'average' will return the relative frequencies, which will be summed to 1.


    NOTE: The ranges parameter will be automatically set to the folloiwng, if not specified explicitly.
          ranges = [[np.min(stimulus_feature[:,0]), np.max(stimulus_feature[:,0])],
                    [np.min(stimulus_feature[:,1]), np.max(stimulus_feature[:,1])]]

    - Return Value Specifications
    figure: The whole resulting figure.
    ax: A plot visualized in the figure.
    """
    if not isinstance(fig_size, tuple):
        raise TypeError(f"fig_size should be a tuple, got the type of {type(fig_size).__name__}.")
    if any([not isinstance(fs, int) for fs in fig_size]):
        raise ValueError(f"fig_size should contain integer elements.")
    if len(fig_size)!=2:
        raise ValueError(f"fig_size should be of length 2, got {len(fig_size)}.")
    if not isinstance(num_data, int):
        raise TypeError(f"num_data should be an integer value, got the type of {type(num_data).__name__}.")
    if not isinstance(stimulus_feature, np.ndarray):
        raise TypeError(f"stimulus_feature should be a numpy array, got the type of {type(stimulus_feature).__name__}.")
    if stimulus_feature.ndim!=2:
        raise ValueError(f"stimulus_feature should be a 2D array, got {stimulus_feature.ndim} dimensions.")
    if stimulus_feature.shape[1]!=2:
        raise ValueError(f"stimulus_feature should have two columns, got {stimulus_feature.shape[1]} columns.")
    if not isinstance(bins, list):
        raise TypeError(f"bins should be a list, got the type of {type(bins).__name__}.")
    if not all([isinstance(b, int) for b in bins]):
        raise TypeError(f"bins should contain integer values.")
    if len(bins)!=2:
        raise ValueError(f"bins should be of length 2, got {len(bins)}.")
    if ranges is not None:
        if not isinstance(ranges, list):
            raise TypeError(f"ranges should be a list, got the type of {type(ranges).__name__}.")
        if not all([isinstance(r, list) for r in ranges]):
            raise TypeError(f"ranges should contain list elements.")
        if not all([isinstance(r1, float) for r1 in ranges[0]]):
            raise TypeError(f"ranges[0] should contain float type elements.")
        if not all([isinstance(r2, float) for r2 in ranges[1]]):
            raise TypeError(f"ranges[1] should contain float type elements.")
        if len(ranges)!=2:
            raise ValueError(f"ranges should contain two list elements, got {len(ranges)}.")
        if len(ranges[0])!=2:
            raise ValueError(f"ranges[0] should contain two float elements, got {len(ranges[0])}.")
        if len(ranges[1])!=2:
            raise ValueError(f"ranges[1] should contain two float elements, got {len(ranges[1])}.")
    
    if not isinstance(xlabel, str):
        raise TypeError(f"xlabel should be a string value, got the type of {type(xlabel).__name__}.")
    if not isinstance(ylabel, str):
        raise TypeError(f"ylabel should be a string value, got the type of {type(ylabel).__name__}.")
    if not isinstance(zlabel, str):
        raise TypeError(f"zlabel should be a string value, got the type of {type(zlabel).__name__}.")
    if not isinstance(title, str):
        raise TypeError(f"title should be a string value, got the type of {type(title).__name__}.")
    if not isinstance(mode, str):
        raise TypeError(f"mode should be a string value, got the type of {type(mode).__name__}.")
    if mode not in ["average", "sum"]:
        raise ValueError(f"mode should be either 'average' or 'sum', got {mode}.")

       
    figure=plt.figure()
    ax=figure.add_subplot(projection='3d')
    hist, dv1_edge, dv2_edge=np.histogram2d(stimulus_feature[:,0], stimulus_feature[:,1], bins=bins, range=ranges)
    if mode=="average":
        hist=hist/num_data

    dv1_pos, dv2_pos=np.meshgrid(dv1_edge[:-1], dv2_edge[:-1], indexing="ij")
    dv1_pos=dv1_pos.ravel()
    dv2_pos=dv2_pos.ravel()
    hist_pos=0

    bin_width=(ranges[0][1]-ranges[0][0])/bins[0] if ranges else dv1_edge[1] - dv1_edge[0]
    bin_depth=(ranges[1][1]-ranges[1][0])/bins[1] if ranges else dv2_edge[1] - dv1_edge[0]
    h=hist.ravel()
    
    bin_centers_dv1 = dv1_pos[:-1] + bin_width/2
    bin_centers_dv2 = dv2_pos[:-1] + bin_depth/2
    
    
    ax.bar3d(x=bin_centers_dv1, y=bin_centers_dv2, z=hist_pos, 
             dx=bin_width, dy=bin_depth, dz=h, zsort='average')
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_zlabel(zlabel)
    ax.set_title(title)

    return figure, ax

def plot_convergence(gp_regressor, x_range, y_range, x_data_points, y_data_points, figure_size=(12, 4), function_colors=["lightgreen", "lightblue", "mediumpurple", "black"]):
    """
        Visualize Gaussian Process (GP) regression convergence.

        This function fits a GaussianProcessRegressor incrementally as more data points are added,
        plots the GP mean function at selected quantiles of the training process, and shows
        the convergence of the mean squared error (MSE) between each trial and the final trial.

        Arguments:
            gp_regressor (GaussianProcessRegressor): 
                An initialized scikit-learn GaussianProcessRegressor object.
                Raises TypeError if the object is not of this type.
            x_range (array-like): 
                Full range of possible X values.
            y_range (array-like): 
                Full range of possible Y values.
            x_data_points (array-like): 
                Observed X data points.
            y_data_points (array-like): 
                Observed Y data points.
            figure_size (tuple, optional): 
                Size of the figure as (width, height). Defaults to (12, 4).
            function_colors (list of strings, optional): 
                Colors for visualizing GP mean functions at different quantiles. 
                Defaults to ["lightgreen", "lightblue", "mediumpurple", "black"].

        Returns:
            tuple:
                - figure (matplotlib.figure.Figure): The overall figure object.
                - axes (numpy.ndarray of matplotlib.axes.Axes): Array containing the two subplot axes.

        Raises:
            TypeError: If `gp_regressor` is not an instance of 
                    sklearn.gaussian_process.GaussianProcessRegressor.
            ValueError: If the lengths of `x_data_points` and `y_data_points` do not match
    """

    if not isinstance(gp_regressor, GaussianProcessRegressor):
        raise TypeError(
            f"gp_regressor must be sklearn.gaussian_process.GaussianProcessRegressor, "
            f"got {type(gp_regressor).__name__}."
        )
    
    if len(x_data_points) != len(y_data_points):
        raise ValueError(
            f"x_data_points and y_data_points must have the same length, "
            f"got {len(x_data_points)} and {len(y_data_points)} for x and y respectively."
        )

    num_data_points = len(x_data_points)
        
    figure, axes = plt.subplots(1, 2, figsize=figure_size)
    gp_mean_function_list = []

    quantiles_visualize = [0.25, 0.5, 0.75, 1]
    n_trials_visualize = []
    for quantile in quantiles_visualize:
        n_trials_visualize.append(int(num_data_points * quantile))

    # Generate plot 1: GP mean functions visualized
    x_range_reshaped_for_gpr = np.array(x_range).reshape(-1, 1)
    quantile_count = 0
    for i in range(1, num_data_points+1):
        x_data_points_current_trial = x_data_points[:i]
        x_data_points_reshaped_for_gpr = np.array(x_data_points_current_trial).reshape(-1, 1)

        y_data_points_current_trial = y_data_points[:i]
        y_data_points_reshaped_for_gpr = np.array(y_data_points_current_trial).reshape(-1, 1)
        
        gp_regressor.fit(x_data_points_reshaped_for_gpr, y_data_points_reshaped_for_gpr)

        gp_mean_function = gp_regressor.predict(x_range_reshaped_for_gpr)

        gp_mean_function_list.append(gp_mean_function)
        
        if i in n_trials_visualize[:-1]:
            current_quantile = int(quantiles_visualize[quantile_count] * 100)
            axes[0].plot(x_range, gp_mean_function, color=function_colors[quantile_count], linewidth=2, label=f"Trial #{i} ({current_quantile}%)")
            quantile_count += 1
        elif i == n_trials_visualize[-1]:
            axes[0].plot(x_range, gp_mean_function, color=function_colors[-1], linewidth=2.5, label="Final Trial")


    # Set plot margins for better visibility
    ymin, ymax = min(y_range), max(y_range)
    span = ymax - ymin
    margin = 0.1 * span
    axes[0].set_ylim(ymin - margin, ymax + margin)
    axes[0].set_title("GP mean functions visualized")
    axes[0].legend()

    # Data points scatter
    axes[0].scatter(x_data_points, y_data_points, c="darkgrey", edgecolor="white", zorder=3, s=30)

    # Generate plot 2: MSE value between each trial and the final trial
    mse_values = []
    print(num_data_points)
    print(len(gp_mean_function_list))
    for i in range(num_data_points):
        mse = mean_squared_error(gp_mean_function_list[i], gp_mean_function_list[-1])
        mse_values.append(mse)

    trials = np.arange(1, len(mse_values) + 1) 
    axes[1].plot(trials, mse_values, marker='o', linewidth=2)
    axes[1].xaxis.set_major_locator(MultipleLocator(1))
    axes[1].set_xlim(0.5, len(mse_values)+0.5)
    axes[1].set_xlabel("Trial")
    axes[1].set_ylabel("MSE")
    axes[1].set_title("Distance to the final function")
    axes[1].grid(True, alpha=0.3)

    figure.tight_layout()
    figure.suptitle("Convergence Plot", fontsize=16, y=1.03)

    return figure, axes


from .gpal_optimize import gpal_optimize
from .gpr_instance import GPRInstance
from .utils import argsConstructor, sequence_with_interval, grid_with_sequences
from .gpal_plot import plot_convergence, plot_frequency_histogram_1D, plot_frequency_histogram_2D, plot_GPAL_compare_uncertainty, plot_GPAL_uncertainty

__all__ = ['gpal_optimize', 
           'GPRInstance', 
           'argsConstructor',
           'sequence_with_interval',
           'grid_with_sequences',
           'plot_convergence',
           'plot_frequency_histogram_1D',
           'plot_frequency_histogram_2D',
           'plot_GPAL_compare_uncertainty',
           'plot_GPAL_uncertainty']


from sklearn.gaussian_process.kernels import Kernel
from sklearn.gaussian_process import GaussianProcessRegressor
import numpy as np
import numpy.typing as npt
from functools import wraps
from typing import NamedTuple, Optional, Union, Tuple, Callable, List
from scipy.stats import norm


BoundsType=Union[Tuple[float, float], str]
ScalesType=Union[npt.NDArray[np.float64], float]
npFuncType=Callable[[npt.NDArray, npt.NDArray], npt.NDArray]
MetricsType=Union[str, npFuncType]
kernelTypeDic={0:'ConstantKernel', 
                1:'DotProduct', 
                2:'ExpSineSquared', 
                3:'Exponentiation',
                4:'Matern', 
                5:'PairwiseKernel', 
                6:'RBF', 
                7:'RationalQuadratic', 
                8:'WhiteKernel'}

class prediction(NamedTuple):
    mean: np.ndarray
    std: Optional[np.ndarray] = None
    cov: Optional[np.ndarray] = None

def check_BoundsType_arguments(arg:BoundsType, tuple_elem_type=float, bounds_min=0.0, allowed_strs=["fixed"]):
    if isinstance(arg, tuple):
        if len(arg)!=2:
            raise ValueError(f"arg should be of length 2, got the length of {len(value)}.")
        if not (isinstance(arg[0], tuple_elem_type) and isinstance(arg[1], tuple_elem_type)):
            raise TypeError(f"arg should contain elements of type {tuple_elem_type}.")
        if not (arg[0]>=bounds_min and arg[1]>=bounds_min):
            raise ValueError(f"arg should contain elements not smaller than {bounds_min}.")
        
    elif isinstance(arg, str):
        if arg not in allowed_strs:
            raise ValueError(f"The allowed string values for arg are {allowed_strs}, got {arg}.")
        
    else:
        raise TypeError(f"arg should be either a tuple or a string, got the type of {type(value).__name__}.")
        
def check_ConstantKernel_arguments(arg1:float=1.0, arg2:BoundsType=(1e-5, 1e5)):
    if not isinstance(arg1, float):
        raise TypeError(f"arg1 should be a float value, got the type of {type(arg1).__name__}.")

    check_BoundsType_arguments(arg2)

    return {'constant_value':arg1, 'constant_value_bounds':arg2}


def check_DotProduct_arguments(arg1:float=1.0, arg2:BoundsType=(1e-5, 1e5)):
    if not isinstance(arg1, float):
        raise TypeError(f"arg1 should be a float value, got the type of {type(arg1).__name__}.")
    if not arg1>=0:
        raise ValueError(f"arg1 should be a non-negative float value, got {arg1}.")

    check_BoundsType_arguments(arg2)
    
    return {'sigma_0':arg1, 'sigma_0_bounds':arg2}


def check_ExpSineSquared_arguments(arg1:float=1.0, arg2:float=1.0, arg3:BoundsType=(1e-5, 1e5), arg4:BoundsType=(1e-5, 1e5)):
    if not isinstance(arg1, float):
        raise TypeError(f"arg1 should be a float value, got the type of {type(arg1).__name__}.")
    if not arg1>0:
        raise ValueError(f"arg1 should be a positive value, got {arg1}.")
    
    if not isinstance(arg2, float):
        raise TypeError(f"arg2 should be a float value, got the type of {type(arg1).__name__}.")
    if not arg2>0:
        raise ValueError(f"arg2 should be a positive value, got {arg2}.")
    
    check_BoundsType_arguments(arg3)
    check_BoundsType_arguments(arg4)

    return {'length_scale':arg1, 'periodicity':arg2, 'length_scale_bounds':arg3, 'periodicity_bounds':arg4}


def check_Exponentiation_arguments(arg1:Kernel, arg2:float):
    if not isinstance(arg1, Kernel):
        raise TypeError(f"arg1 should be a valid kernel object, got the type of {type(arg1).__name__}.")
    if not isinstance(arg2, float):
        raise TypeError(f"arg2 should be a float value, got the type of {arg2}.")
    
    return {'kernel':arg1, 'exponent':arg2}


def check_Matern_arguments(arg1:ScalesType=1.0, arg2:BoundsType=(1e-5, 1e5), arg3:float=1.5):
    if not (isinstance(arg1, np.ndarray) or isinstance(arg1, float)):
        raise TypeError(f"arg1 should be either a 1D numpy array or a float value, got the type of {type(arg1).__name__}.")
    elif isinstance(arg1, np.ndarray):
        if arg1.ndim!=1:
            raise ValueError(f"arg1 array should be a 1D numpy array, got {arg1.ndim} dimensions.")
        if arg1.dtype!=np.float64:
            raise TypeError(f"The dtype of arg1 array should be np.float64, got the dtype of {arg1.dtype}.")
        if np.min(arg1) <=0:
            raise ValueError(f"arg1 array should only contain positive elements, got {np.min(arg1)}.")
    else:
        if arg1<=0:
            raise ValueError(f"arg1 should be a positive value, got {arg1}.")

    check_BoundsType_arguments(arg2)

    if not isinstance(arg3, float):
        raise TypeError(f"arg3 should be a float value, got the type of {type(arg3).__name__}.")
    if arg3<=0:
        raise ValueError(f"arg3 should be a positive value, got {arg3}.")
    

    return {'length_scale':arg1, 'length_scale_bounds':arg2, 'nu': arg3}            

'''
def check_PairwiseKernel_argument_3(arg3:npFuncType):
    @wraps(arg3)
    def check_pairwise_arguments(data1:npt.NDArray, data2:Optional[npt.NDArray]):
        if not isinstance(data1, np.ndarray):
            raise TypeError(f"The first argument of arg3 function should be a numpy array, got the type of {type(data1).__name__}.")
        if data1.ndim!=2:
            raise ValueError(f"The first argument of arg3 function should be a 2D numpy array, got {data1.ndim} dimensions.")
        if data2 is not None:
            if not isinstance(data2, np.ndarray):
                raise TypeError(f"The second argument of arg3 function should be a numpy array, got the type of {type(data2).__name__}.")
            if data2.ndim!=2:
                raise ValueError(f"The second argument of arg3 function should be a 2D numpy array, got {data2.ndim} dimensions.")
            if data1.shape[1]!=data2.shape[1]:
                raise ValueError(f"The two arguments of arg3 functon should have equal number of columns, got {data1.shape[1]} and {data2.shape[1]}.")
            result=arg3(data1, data2)
        else:
            result=arg3(data1, data1)
        if not isinstance(result, np.ndarray):
            raise TypeError(f"The return value of arg3 function should be a numpy array.")
        if result.ndim!=2 or result.shape[0]!=result.shape[1]:
            raise ValueError(f"The return value of the arg3 callable should be a 2D square numpy array, got the shape of {result.shape}.")
        if not np.allclose(result, result.T, atol=1e-10):
            raise ValueError(f"The return value of the arg3 callable should be symmetric.")
        try:
            np.linalg.cholesky(result+1e-10*np.eye(result.shape[0]))
        except np.linalg.LinAlgError:
            raise ValueError("The return value of arg3 callable should be a positive semi-definite kernel matrix.")
        return None
    return checker

'''

## arg4: dict[parameter : argument]
## param_type_dic: dict[parameter : type]
## param_default_dic: dict[parameter : defaults]
def check_PairwiseKernel_argument_4(arg4:dict, param_type_dic:dict, param_default_dic:dict):
    ## Empty dictionary - for additive_chi2 kernel.
    if param_type_dic=={}:
        return {}
    
    all_params_list=list(param_type_dic.keys())
    all_params_list.sort()
    all_params_set=set(all_params_list)

    all_input_params_list=list(arg4.keys())
    input_param_indices_set=set([all_params_list.index(param) for param in all_input_params_list])
    
    if not input_param_indices_set.issubset(all_params_set):
        raise ValueError(f'''The input dictionary arg4 should contain keys included in the following list: {all_params_list}.\n 
                             Note that not all available keys need to be present.''')
    
    for input_param in all_input_params_list:
        input_argument=arg4[input_param] # Argument of interest
        expected_type=param_type_dic[input_param] # Expected type of aoi value
        if not isinstance(input_argument, expected_type):
            raise TypeError(f"The arg4[{input_param}] should be of type {expected_type}, got the type of {type(input_argument).__name__}")
    
    for param in all_params_list:
        if param not in all_input_params_list:
            arg4[param]=param_default_dic[param]

    return arg4

def check_PairwiseKernel_arguments(arg1:float=1.0, arg2:BoundsType=(1e-5, 1e5), arg3:MetricsType="linear", arg4:Optional[dict]=None):
    if not isinstance(arg1, float):
        raise TypeError(f"arg1 should be a float value, got the type of {type(arg1).__name__}.")
    else:
        if arg1<=0:
            raise ValueError(f"arg1 should be a positive value, got {arg1}.")
    check_BoundsType_arguments(arg2)

    arg3Cands=["linear", "additive_chi2", "chi2", "poly", "polynomial", "rbf", "laplacian", "sigmoid", "cosine", "precomputed"]
    

    linear_kwargs_type={'dense_output': bool}
    additive_chi2_kwargs_type={}
    chi2_kwargs_type={'gamma': float}
    poly_kwargs_type={'degree': float, 'gamma': Optional[float], 'coef0': float}
    rbf_kwargs_type={'gamma': Optional[float]}
    laplacian_kwargs_type={'gamma': Optional[float]}
    sigmoid_kwargs_type={'gamma': Optional[float], 'coef0': float}
    cosine_kwargs_type={'dense_output': bool}

    linear_kwargs={'dense_output': True}
    additive_chi2_kwargs={}
    chi2_kwargs={'gamma': 1.0}
    poly_kwargs={'degree': 3.0, 'gamma': None, 'coef0': 1.0}
    rbf_kwargs={'gamma': None}
    laplacian_kwargs={'gamma': None}
    sigmoid_kwargs={'gamma': None, 'coef0': 1.0}
    cosine_kwargs={'dense_output': True}
    
    if not (isinstance(arg3, str) or isinstance(arg3, Callable)):
        raise TypeError(f"arg3 should be a Callable or a string, got the type of {type(arg3).__name__}")

    if isinstance(arg3, str):
        if arg3 not in arg3Cands:
            raise ValueError(f"arg3 string value should be one of the following values: {arg3Cands}")
        ## For the case wehre arg3=="precomputed",
        ## sklearn.metrics.pairwise automatically checks
        ## whether the input array is symmetric
        ## and therefore able to be regarded as a kernel matrix.
        ## So whatever we pass does not matter, as long as the types match.
        if arg3=="precomputed":
            return {'gamma':arg1, 'gamma_bounds':arg2, 'metric':arg3, 'pairwise_kernel_args':None}
    ## For the case where arg3 is a Callable object
    ## there are no restrictions on the dimension of input array.
    '''
    if isinstance(arg3, Callable):
        pairwiseArg3(arg3)
    '''

    if arg4 is not None:
        if not isinstance(arg4, dict):
            raise TypeError(f"arg4 should be a dictionary or None, got the type of {type(arg4).__name__}")
        else:
            if arg3=="linear":
                type_dic=linear_kwargs_type
                default_dic=linear_kwargs
            elif arg3=="additive_chi2":
                type_dic=additive_chi2_kwargs_type
                default_dic=additive_chi2_kwargs
            elif arg3=="chi2":
                type_dic=chi2_kwargs_type
                default_dic=chi2_kwargs
            elif arg3=="poly" or arg3=="polynomial":
                type_dic=poly_kwargs_type
                default_dic=poly_kwargs
            elif arg3=="rbf":
                type_dic=rbf_kwargs_type
                default_dic=rbf_kwargs
            elif arg3=="laplacian":
                type_dic=laplacian_kwargs_type
                default_dic=laplacian_kwargs
            elif arg3=="sigmoid":
                type_dic=sigmoid_kwargs_type
                default_dic=sigmoid_kwargs
            elif arg3=="cosine":
                type_dic=cosine_kwargs_type
                default_dic=cosine_kwargs
            else:
                raise ValueError(f"arg3 string value should be one of the following options: {arg3Cands}")
            
            kwargs=check_PairwiseKernel_argument_4(arg4=arg4, 
                                                   param_type_dic=type_dic, 
                                                   param_default_dic=default_dic)

            if arg3=="laplacian":
                if "gamma" in list(kwargs.keys()):
                    gamma=kwargs['gamma']
                    if gamma <=0:
                        raise ValueError(f"arg4['gamma'] should be a positive value, got {gamma}.")
    
    return {'gamma':arg1, 'gamma_bounds':arg2, 'metric':arg3, 'pairwise_kernel_kwargs':kwargs}
    

def check_RBF_arguments(arg1:ScalesType=1.0, arg2:BoundsType=(1e-05, 1e5)):
    if not (isinstance(arg1, np.ndarray) or isinstance(arg1, float)):
        raise TypeError(f"arg1 should be a 1D numpy array or a float value, got the type of{type(arg1).__name__}.")
    if isinstance(arg1, np.ndarray):
        if arg1.ndim!=1:
            raise ValueError(f"arg1 array should be 1D, got {arg1.ndim} dimensions.")
        if arg1.dtype!=np.float64:
            raise TypeError(f"The dtype of arg1 should be np.float64, got {arg1.dtype}.")
        if np.min(arg1)<=0:
            raise ValueError(f"arg1 array should only contain positive elements, got {np.min(arg1)}.")
    else:
        if arg1<=0:
            raise ValueError(f"arg1 should be a positive float value, got {arg1}.")
    check_BoundsType_arguments(arg2)
    
    return {'length_scale':arg1, 'length_scale_bounds':arg2}                


def check_RationalQuadratic_arguments(arg1:float=1.0, arg2:float=1.0, arg3:BoundsType=(1e-5,1e5), arg4:BoundsType=(1e-5,1e5)):
    if not isinstance(arg1, float):
        raise TypeError(f"arg1 should be a float value, got the type of {type(arg1).__name__}.")
    else:
        if arg1<=0:
            raise ValueError(f"arg1 should be a positive value, got {arg1}.")
    
    if not isinstance(arg2, float):
        raise TypeError(f"arg2 should be a float value, got the type of {type(arg2).__name__}.")
    else:
        if arg2<=0:
            raise ValueError(f"arg2 should be a positive value, got {arg2}.")
    
    check_BoundsType_arguments(arg3)
    check_BoundsType_arguments(arg4)

    return {'length_scale':arg1, 'alpha':arg2, 'length_scale_bounds':arg3, 'alpha_bounds':arg4}


def check_WhiteKernel_arguments(arg1:float=1.0, arg2:BoundsType=(1e-5,1e5)):
    if not isinstance(arg1, float):
        raise TypeError(f"arg1 should be a float value, got the type of {type(arg1).__name__}.")
    else:
        if arg1<=0:
            raise ValueError(f"arg1 should be a positive value, got {arg1}.")

    check_BoundsType_arguments(arg2)

    return {'noise_level':arg1, 'noise_level_bounds':arg2}


    
def argsConstructor(kernel_type_list:List, kernel_arguments_list:List[List]):
    if not isinstance(kernel_type_list, list):
        raise TypeError(f"kernel_type_list should be a list, got the type of {type(kernel_type_list).__name__}.")
    if not all(isinstance(kt, int|str) for kt in kernel_type_list):
        typ=kernel_type_list[[isinstance(kt, int|str) for kt in kernel_type_list].index(False)]
        raise TypeError(f"kernel_type_list should contain either integer or tring values, got the type of {type(typ).__name__ }.")
    num_kernels=len(kernel_type_list)
    if not isinstance(kernel_arguments_list, list):
        raise TypeError(f"kernel_arguments_list should be a list, got the type of {type(kernel_arguments_list).__name__}.")
    if not all(isinstance(kal, list) for kal in kernel_arguments_list):
        typ=kernel_arguments_list[[isinstance(kal, list) for kal in kernel_arguments_list].index(False)]
        raise TypeError(f"kernel_arguments_list should contain list elements, got the type of {type(typ).__name__ }.")
    if len(kernel_arguments_list)!=num_kernels:
        raise ValueError(f"kernel_arguments_list should be of length {num_kernels}, got {len([kernel_arguments_list])}.")
    
    kernel_type_index_list=[i for i in range(9)]
    kernel_type_names_list=["ConstantKernel", "DotProduct", "ExpSineSquared", "Exponentiation",
                            "Matern", "PairwiseKernel", "RBF", "RationalQuadratic", "WhiteKernel"]
    kernel_type_options_list=kernel_type_index_list+kernel_type_names_list
    kti_masks=[kt not in kernel_type_options_list for kt in kernel_type_list]
    if any(kti_masks):
        raise ValueError(f"The kernel type should be the following: \n An integer index within (0,9) or the name of the valid kernel types.")
    
    kernel_types=[kernelTypeDic[kt] for kt in kernel_type_list if isinstance(kt, int)]
    kernel_arguments_dict_list=[]

    for kernel_index, kernel_type in enumerate(kernel_types):
        args=kernel_arguments_list[kernel_index]
        if kernel_type=="ConstantKernel":
            args_dict=check_ConstantKernel_arguments(*args)
        elif kernel_type=="DotProduct":
            args_dict=check_DotProduct_arguments(*args)
        elif kernel_type=="ExpSineSquared":
            args_dict=check_ExpSineSquared_arguments(*args)
        elif kernel_type=="Exponentiation":
            args_dict=check_Exponentiation_arguments(*args)
        elif kernel_type=="Matern":
            args_dict=check_Matern_arguments(*args)
        elif kernel_type=="PairwiseKernel":
            args_dict=check_PairwiseKernel_arguments(*args)
        elif kernel_type=="RBF":
            args_dict=check_RBF_arguments(*args)
        elif kernel_type=="RationalQuadratic":
            args_dict=check_RationalQuadratic_arguments(*args)
        elif kernel_type=="WhiteKernel":
            args_dict=check_WhiteKernel_arguments(*args)
        else:
            raise ValueError(f"The following kernel type is not valid: {kernel_type}.")
        kernel_arguments_dict_list.append(args_dict)
    
    return kernel_types, kernel_arguments_dict_list

def sequence_with_interval(start_val: int|float, end_val: int|float, interval:int|float):
    if not isinstance(start_val, float|int):
        raise TypeError(f"start_val should be a float or an integer value, got the type of {type(start_val).__name__}.")
    if not isinstance(end_val, float|int):
        raise TypeError(f"end_val should be a float or an integer value, got the type of {type(end_val).__name__}.")
    if not isinstance(interval, int|float):
        raise TypeError(f"interval should be a float or an integer value, got the type of {type(interval).__name__}.")
    if interval==0:
        raise ValueError(f"The specified interval is zero; this will result in an infinite sequence.")
    if (end_val - start_val)*interval<0:
        raise ValueError(f"Wrong signs; the sequence starts from {start_val} to {end_val}, with an interval of {interval}.")

    return np.arange(start_val, end_val, interval, dtype=float).reshape(-1, 1)

def grid_with_sequences(*sequences):
    seq_2D= (lambda *args: np.concat(*args, axis=0))(*sequences)
    coords_per_axis=list(np.meshgrid(*seq_2D, indexing='ij'))
    coords_grid=np.stack(coords_per_axis, -1)
    return coords_grid




        










## Importing required Python packages
import numpy as np
import pandas as pd
import os, random, warnings

## Importing key functions from our gpal package.
## These three functions must be utilized to conduct GPAL properly.
from gpalexp import GPRInstance, gpal_optimize, argsConstructor
from gpalexp import sequence_with_interval

## Importing key functions from the psychopy package.
## Our number-line task file is implemented based on the psychopy package.
from psychopy import event, logging
## Importing functions for settings of number-line task, from nlt_setup.py file.
from nlt_setup import show_and_get_response, initialize_psychopy

## Managing default settings to ignore warning messages raised by psychopy.
warnings.filterwarnings('ignore')
logging.console.setLevel(logging.ERROR)


## Initializing visual elements for our number-line task code.
## initialize_psychopy(): setting up the psychopy visual elements.
## NOTE: The 'fullscreen mode' is turned off by default.
##       Users can turn on the fullscreen mode by setting fullscr=True.
visuals = initialize_psychopy(fullscr=False)



'''========================================== Step 0 =========================================='''

'''
Defining a Gaussian process regressor (GPR) object.
'''
## argsConstructor() is a function which generates values to create a GPR object.
## argsConsturctor() should take 3 values: num_kernels, kernel_type_list, kernel_arguments_list
## num_kernels indicate the number of kernel objects to be combined,
## which exactly corresponds to what n_kernel argument value represents.
## kernel_type_list should be a list object containing the types (or their indices) of kernel objects to be combined.
## It is already loaded in kernel_types, based on line 55 of this file.
## kernel_arguments_list should be a list object holding the values to be specified to properly create each kernel objects.
## We've loaded that list object in kernel_arguments in line 56.
## Therefore, we can use the argsConsturctor() function in the following way.

## argsConsturctor() has two outputs, which we've named kernel_type and kernel_args.
## We will soon exploit these outputs when generating a GPR object.
## NOTE: It is sufficient to write only the values we are putting to argsConsturctor(),
##       But for guidance, we've specified both the values that argsConstructor() should take
##       and those we've loaded and putting into the function. 
kernel_type, kernel_args = argsConstructor([0,6,8], [[1.0], [1.0], [0.01]])


'''
Initializing a kernel object and a GPR with the kernel.
'''
## For GPAL, we need to create a GPR object.
## In this package, GPRInstance() function takes the role.

## GPRInstance() takes 7 values, but the last one is optional and need not to be specified.
## kernel_types should have a list object holding the types of each kernel objects to be combined,
## which we've just created with argsConstructor() and named kernel_type.
## Therefore, we can just provide kernel_type for kernel_types value.
## Similarly, it is sufficient to provide kernel_args for kernel_arguments value.

## combine_format should be a string object, which specifies the way those individual kernel objects are combined.
## In combine_format, we represent each kernel object as k1, k2, k3, and so on.
## For example, if we have 3 kernels and we want to multiply the first two ones and add the last one,
## we can just feed "k1*k2+k3" as combine_format.

## For alpha, n_restarts_optimizer, normalize_y, and random_state, 
## it is recommended to put the default values, which are automatically fed into.

## There are two outputs, which we've named kernel and gpr.
## kernel is a Gaussian process kernel object created following our specifications.
## gpr is a GPR object associated with that kernel object.
combine_format = "k1*k2+k3"
kernel, gpr = GPRInstance(kernel_type, kernel_args, combine_format)

''' =================================== Step 1 ========================================='''

num_trials=20           # Number of experiment trials for a single subject.
num_features=1          # The number of stimulus feautres to be optimized.

''' 
Initializing a numpy array for recording. 
'''
## The number-line task of our interest is a 1-dimensional task,
## where the 'given number' may vary but 'upper bound' stays still.
## In other words, we have a single stimulus feature, which is the 'given number'.
## We will create a numpy array for recording the experiment results.
## The first row will record the 'given number' for each trial,
## and the second row is for recording the subject's estimation on the 'given number'. 
## This record array will be updated after each trial.
## NOTE: The number of columns of record_array is set to num_features+1. 
##       Here num_features is the number of stimulus features to be optimized.
##       This enables us to record multiple stimulus features and user responses, 
#        in a single data structure, even for arbitrary number of stimulus features.
##       The first num_features columns will record value of each stimulus features,
##       and the last column will record the subject's responses.
data_record = np.zeros((20, 2))   


''' 
Defining the number-line task-specific values. 
'''
## 1-dimensional number-line task has a constant 'upper bound' value.
## We've set it to 500, and named it 'max_number'
## to refer to the upper bound value with the name of 'max_number' in the following codes.
max_number = 500


## In our number-line task, there are two types of trials.
## One is a 'fixed-sized' trial, where the size of the presented dots remains a default value.
## The other is a 'variable-sized' trial, where the size of the dots may vary.
## size_control_order determines the order of fixed-size trials and variable-sized ones randomly.
dot_size_flags = [True] * (num_trials//2) + [False] * (num_trials//2)
size_control_order = random.sample(dot_size_flags, num_trials)


'''
Running the first trial.
'''
## This code block is for running the first trial.
## Since we cannot optimize the stimulus feature (i.e. 'given number') in the first trial,
## we just set it as a random number among (5, 10, 15, ... , 495, 500)
## pMean, pStd, lml are GPAL-related statistics, which cannot be calculated in the first trial.
## Therefore we've just initialized them with simple values.
start_val=5
end_val=500
interval=5
stimuli=sequence_with_interval(start_val, end_val, interval)

trial_index=0
initial_stimulus=np.random.choice(stimuli.squeeze())
gp_mean = 0
gp_std = 1
lml = 0


## Show the dots and get response from participant
## If size_control is True, the function internally adjusts the size of the dots.
## Otherwise, the default-sized dots are provided.
response = show_and_get_response(initial_stimulus, 
                                 visuals, 
                                 max_number, 
                                 size_control_order[0])

'''
Recording the results for the next trial
'''
## The 0-th row records the selected value of the stimulus feature, namely the 'given number' of the number-line task. 
## The 1-th row records the response of the subject for the given_number.
data_record[trial_index,0] = initial_stimulus
data_record[trial_index,1] = response

## Waiting for the user to press the space key, to move on to the next trial
event.waitKeys(keyList=['space'])  


''' Running experimental trials with GPAL. '''
## We will run the following block n_trials times, with trial_idx representing the index of the currently running trial.
for trial_index in range(1, num_trials):

    ## This code block is executed otherwise (i.e. for the second to the last trial).
    ## gpal_optimize() function actually executes GPAL optimization
    ## and yields an optimal feature for the next trial
    ## as well as some GPAL-related statistics.
    ## NOTE: The stimulus feature to be optimized here is the 'given number' of the number-line task.
    
    
    ## Executing the gpal_optimize() function with appropriate input values.
    optimal_design, pMean, pStd, lml = gpal_optimize(gpr,                                  # A GP regressor object to be fitted.
                                                     num_features,                         # Number of feature stimulus to be optimized
                                                     data_record[:trial_index],            # The feature stimulus data for fitting the GP regressor
                                                     stimuli                               # Overall specifications on the design candidate values.
                                                    )                  
    given_number = int(optimal_design)                                                     # Extracting the optimal 'given number' value for the next trial.


    # Show the dots and get response from participant
    # If size_control is enabled, adjust the dot size, else use default size
    response = show_and_get_response(given_number, visuals, max_number=max_number, size_control=size_control_order[trial_index])

    '''
    Recording the results for the next trial
    '''
    ## The 0-th row records the selected value of the stimulus feature, namely the 'given number' of the number-line task. 
    ## The 1-th row records the response of the subject for the given_number.
    data_record[trial_index,0] = given_number
    data_record[trial_index,1] = response

    ## Waiting for the user to press the space key, to move on to the next trial
    event.waitKeys(keyList=['space'])  

'''
Saving experiment results in the .csv format
'''
save_results_dir='results'
if not os.path.exists(save_results_dir):
    os.mkdir(save_results_dir)
results_df = pd.DataFrame(data_record, columns=['given_number', 'response'])
results_df.to_csv(os.path.join(save_results_dir, f'results_trial_{num_trials}.csv'), index=False)

## Closing the psychopy experiment window.
visuals['win'].close() 